/*
 * Copyright (C) 2013 - 2016 Sony Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _LDACBT_EX_H_
#define _LDACBT_EX_H_

#include "ldacBT.h"

enum {
/* undocumented settings. for internal use only */
    LDACBT_EQMID_Q0 = LDACBT_EQMID_MQ + 1,
    LDACBT_EQMID_Q1,
    LDACBT_EQMID_Q2,
    LDACBT_EQMID_Q3,
    LDACBT_EQMID_Q4,
    LDACBT_EQMID_Q5,
    LDACBT_EQMID_Q6,
    LDACBT_EQMID_Q7,
    LDACBT_EQMID_Q8,
    LDACBT_EQMID_END,
};


#endif /* _LDACBT_EX_H_ */

